// 2008 IDesign Inc. All rights reserved 
// Questions? Comments? Go to http://www.idesign.net

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Transactions;

namespace WCFTransactions
{
	public class Transactional<T> : ISinglePhaseNotification
	{
		public int Port;
		public bool Trace;

		T _value;
		T _temporaryValue;
		Transaction _currentTransaction;
		TransactionalLock _lock;

		public T Value
		{
			get { return GetValue(); }
			set { SetValue(value); }
		}

		public Transactional()
			: this(default(T))
		{
		}

		public Transactional(T value)
		{
			_lock = new TransactionalLock();
			_value = value;
		}

		public Transactional(Transactional<T> transactional)
			: this(transactional.Value)
		{ 
		}

		static Transactional()
		{
			ResouceManager.ConstrainType(typeof(T));
		}

		void ISinglePhaseNotification.SinglePhaseCommit(SinglePhaseEnlistment singlePhaseEnlistment)
		{
			Commit();
			singlePhaseEnlistment.Committed();
		}

		void Commit()
		{
			if(Trace)
				Console.WriteLine("COMMIT {0}", Port);

			IDisposable disposable = _value as IDisposable;

			if(disposable != null)
				disposable.Dispose();
			
			_value = _temporaryValue;
			_currentTransaction = null;
			_temporaryValue = default(T);
			_lock.Unlock();
		}

		void IEnlistmentNotification.Commit(Enlistment enlistment)
		{
			Commit();

			enlistment.Done();
		}

		void IEnlistmentNotification.InDoubt(Enlistment enlistment)
		{
			_lock.Unlock();

			enlistment.Done();
		}

		void IEnlistmentNotification.Prepare(PreparingEnlistment preparingEnlistment)
		{
			if(Trace)
				Console.WriteLine("PREPARE {0}", Port);

			preparingEnlistment.Prepared();
		}

		void IEnlistmentNotification.Rollback(Enlistment enlistment)
		{
			if(Trace)
				Console.WriteLine("ROLLBACK {0}", Port);

			_currentTransaction = null;

			IDisposable disposable = _temporaryValue as IDisposable;

			if(disposable != null)
				disposable.Dispose();

			_temporaryValue = default(T);
			_lock.Unlock();

			enlistment.Done();
		}

		void Enlist(T t)
		{
			Debug.Assert(_currentTransaction == null);
			_currentTransaction = Transaction.Current;

			Debug.Assert(_currentTransaction.TransactionInformation.Status == TransactionStatus.Active);
			_currentTransaction.EnlistVolatile(this, EnlistmentOptions.None);
			
			_temporaryValue = ResouceManager.Clone(t);
		}

		void SetValue(T t)
		{
			_lock.Lock();

			if(_currentTransaction == null)
			{
				if(Transaction.Current == null)
				{
					_value = t;
					return;
				}
				else
				{
					Enlist(t);
					return;
				}
			}
			else
			{
				// Must have acquired the lock
				Debug.Assert(_currentTransaction == Transaction.Current, "Invalid state in the volatile resource state machine");
				_temporaryValue = t;
			}
		}

		T GetValue()
		{
			_lock.Lock();

			if(_currentTransaction == null)
			{
				if(Transaction.Current == null)
					return _value;
				else
					Enlist(_value);
			}
			
			// Must have acquired the lock
			Debug.Assert(_currentTransaction == Transaction.Current, "Invalid state in the volatile resource state machine");
			return _temporaryValue;
		}

		public static implicit operator T(Transactional<T> transactional)
		{
			return transactional.Value;
		}

		public static bool operator ==(Transactional<T> t1, Transactional<T> t2)
		{
			// Is t1 and t2 null (check the value as well).
			bool t1Null = (Object.ReferenceEquals(t1, null) || t1.Value == null);
			bool t2Null = (Object.ReferenceEquals(t2, null) || t2.Value == null);

			// If they are both null,return true.
			if(t1Null && t2Null)
				return true;

			// If one is null,return false.
			if(t1Null || t2Null)
				return false;
			
			return EqualityComparer<T>.Default.Equals(t1.Value, t2.Value);
		}

		public static bool operator ==(Transactional<T> t1, T t2)
		{
			// Is t1 and t2 null (check the value as well).
			bool t1Null = (Object.ReferenceEquals(t1, null) || t1.Value == null);
			bool t2Null = t2 == null;

			// If they are both null,return true.
			if(t1Null && t2Null)
				return true;

			// If one is null,return false.
			if(t1Null || t2Null)
				return false;
			
			return EqualityComparer<T>.Default.Equals(t1.Value, t2);
		}

		public static bool operator ==(T t1, Transactional<T> t2)
		{
			// Is t1 and t2 null (check the value as well)
			bool t1Null = t1 == null;
			bool t2Null = (Object.ReferenceEquals(t2, null) || t2.Value == null);

			// If they are both null,return true.
			if(t1Null && t2Null)
				return true;

			// If one is null,return false.
			if(t1Null || t2Null)
				return false;
			
			return EqualityComparer<T>.Default.Equals(t1, t2.Value);
		}

		public static bool operator !=(T t1, Transactional<T> t2)
		{
			return !(t1 == t2);
		}

		public static bool operator !=(Transactional<T> t1, T t2)
		{
			return !(t1 == t2);
		}

		public static bool operator !=(Transactional<T> t1, Transactional<T> t2)
		{
			return !(t1 == t2);
		}

		public override int GetHashCode()
		{
			return Value.GetHashCode();
		}

		public override bool Equals(object obj)
		{
			return Value.Equals(obj);
		}
	}
}
